@@ -27,6 +27,7 @@ import apt_pkg
 import ConfigParser
 import copy
 import datetime
+import fcntl
 import re
 import os
 import string
@@ -60,6 +61,40 @@ PROGRESS_LOG="/var/run/unattended-upgrades.progress"
 # set from the sigint signal handler
 SIGNAL_STOP_REQUEST=False
 
+class UnattendedUpgradesCache(apt.Cache):
+    
+    def __init__(self, rootdir, allowed_origins):
+        apt.Cache.__init__(self, rootdir=rootdir)
+        self.allowed_origins = allowed_origins
+        # ensure we update the candidate versions 
+        self.adjust_candidate_versions()
+    def clear(self):
+        apt.Cache.clear(self)
+        # ensure we update the candidate versions 
+        self.adjust_candidate_versions()
+    def adjust_candidate_versions(self):
+        """ Adjust candidate versions to match highest allowed origin
+        
+            This adjusts the origin even if the candidate has a higher
+            version
+        """
+        for pkg in self:
+            # important! this avoids downgrades below
+            if not pkg.is_upgradable:
+                continue
+            # check if we have a version in a allowed origin that is
+            # not the candidate
+            new_cand = None
+            for ver in pkg.versions:
+                if is_allowed_origin(ver, self.allowed_origins):
+                    # leave as soon as we have the highest new candidate
+                    new_cand = ver
+                    break
+            if new_cand and new_cand != pkg.candidate:
+                logging.debug("adjusting candidate version: '%s'" % new_cand)
+                pkg.candidate = new_cand
+
+
 class LogInstallProgress(apt.progress.base.InstallProgress):
     """ Install progress that writes to self.LOG
         (/var/run/unattended-upgrades.progress by default)
@@ -71,7 +106,46 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
         f=open(self.LOG, "w")
         f.write(_("Progress: %s %% (%s)") % (percent, pkg))
         f.close()
-        
+
+    def _fixup_fds(self):
+        required_fds = [ 0, 1, 2, # stdin, stdout, stderr
+                         self.writefd,
+                         self.write_stream.fileno(),
+                         self.statusfd,
+                         self.status_stream.fileno()
+                       ]
+        # ensure that our required fds close on exec
+        for fd in required_fds[3:]:
+            old_flags = fcntl.fcntl(fd, fcntl.F_GETFD)
+            fcntl.fcntl(fd, fcntl.F_SETFD, old_flags | fcntl.FD_CLOEXEC)
+        # close all fds
+        proc_fd = "/proc/self/fd"
+        if os.path.exists(proc_fd):
+            error_count = 0
+            for fdname in os.listdir(proc_fd):
+                try:
+                    fd = int(fdname)
+                except Exception as e:
+                    print "ERROR: can not get fd for '%s'" % fdname
+                if fd in required_fds:
+                    continue
+                try:
+                    os.close(fd)
+                    #print "closed: ", fd
+                except OSError as e:
+                    # there will be one fd that can not be closed
+                    # as its the fd from pythons internal diropen()
+                    # so its ok to ignore one close error
+                    error_count += 1
+                    if error_count > 1:
+                        print "ERROR: os.close(%s): %s" % (fd, e)
+
+    def fork(self):
+        pid = os.fork()
+        if pid == 0:
+            self._fixup_fds()
+        return pid
+
 
 class Unlocked:
     """ context manager for unlocking the apt lock while cache.commit()
@@ -130,6 +204,9 @@ def get_allowed_origins_legacy():
             (distro_id, distro_codename) = s.split(':')
         else:
             (distro_id, distro_codename) = s.split()
+        # escape "," (see LP: #824856) - i wonder if there is a simpler way?
+        distro_id = re.sub(r'([^\\]),', r'\1\\,', distro_id)
+        distro_codename = re.sub(r'([^\\]),', r'\1\\,', distro_codename)
         # convert to new format
         allowed_origins.append("o=%s,a=%s" % (substitute(distro_id), 
                                               substitute(distro_codename)))
@@ -196,6 +273,10 @@ def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg):
     return res
 
 def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
+
+    install_log = LogInstallProgress()
+    install_log.LOG += ".minimal-steps"
+
     # setup signal handler
     signal.signal(signal.SIGUSR1, signal_handler)
 
@@ -226,15 +307,17 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
             cache.clear()
 
         # write progress log information
-        f=open(PROGRESS_LOG+".minimal-steps", "w")
-        percent = len(to_upgrade)/len(pkgs_to_upgrade)*100.0
-        f.write(_("Progress: %s %%: (%s)") % (
-                    percent, ",".join(smallest_partition)))
-        f.close()
-
+        if len(pkgs_to_upgrade) > 0:
+            percent = (len(pkgs_to_upgrade)-len(to_upgrade)) / float(len(pkgs_to_upgrade))*100.0
+        else:
+            percent = 100.0
+        install_log.status_change(pkg=",".join(smallest_partition),
+                                  percent=percent,
+                                  status="")
         # apply changes
         logging.debug("applying set %s" % smallest_partition)
         rewind_cache(cache, [cache[name] for name in smallest_partition])
+
         try:
             iprogress = LogInstallProgress()
             with Unlocked():
@@ -255,10 +338,10 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
             break
     return True
 
-def is_allowed_origin(pkg, allowed_origins):
-    if not pkg.candidate:
+def is_allowed_origin(ver, allowed_origins):
+    if not ver:
         return False
-    for origin in pkg.candidate.origins:
+    for origin in ver.origins:
         for allowed in allowed_origins:
             if match_whitelist_string(allowed, origin):
                 return True
@@ -272,7 +355,7 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist):
             logging.debug("pkg '%s' now marked delete" % pkg.name)
             return False
         if pkg.marked_install or pkg.marked_upgrade:
-            if not is_allowed_origin(pkg, allowed_origins):
+            if not is_allowed_origin(pkg.candidate, allowed_origins):
                 logging.debug("pkg '%s' not in allowed origin" % pkg.name)
                 return False
             for blacklist_regexp in blacklist:
@@ -328,6 +411,10 @@ def conffile_prompt(destFile, prefix=""):
                     # ignore if conffile is obsolete or does not exist
                     if obs == "obsolete" or not os.path.exists(prefix+conf_file):
                         continue
+                    # ignore state "newconffile" until its clearer if there
+                    # might be a dpkg prompt (LP: #936870)
+                    if md5 == "newconffile":
+                        continue
                     # get conffile value from pkg, its ok if the new version
                     # does not have conffiles anymore
                     deb = apt_inst.DebFile(destFile)
@@ -412,7 +499,7 @@ def setup_apt_listchanges(conf="/etc/apt/listchanges.conf"):
     os.environ["APT_LISTCHANGES_FRONTEND"] = "none"
 
 def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
-    " send mail (if configured in Unattended-Upgrades::Mail) "
+    " send mail (if configured in Unattended-Upgrade::Mail) "
     email = apt_pkg.config.find("Unattended-Upgrade::Mail", "")
     if not email:
         return
@@ -444,16 +531,63 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
         s += " " + wrap(" ".join(pkgs_kept_back), 70, " ")
         s += "\n"
     s += "\n"
-    s += _("Package installation log:")+"\n"
-    s += open(logfile_dpkg).read()
-    s += "\n\n"
+    if os.path.exists(logfile_dpkg):
+        s += _("Package installation log:")+"\n"
+        s += open(logfile_dpkg).read()
+        s += "\n\n"
     s += _("Unattended-upgrades log:\n")
     s += mem_log.getvalue()
     mail.stdin.write(s)
     mail.stdin.close()
     ret = mail.wait()
     logging.debug("mail returned: %s" % ret)
+
+
+def do_install(cache, pkgs_to_upgrade, options, logfile_dpkg):
+    # set debconf to NON_INTERACTIVE, redirect output
+    os.putenv("DEBIAN_FRONTEND","noninteractive");
+    setup_apt_listchanges()
     
+    # redirect to log
+    REDIRECT_INPUT = os.devnull
+    fd = os.open(REDIRECT_INPUT, os.O_RDWR)
+    os.dup2(fd,0)
+
+    logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
+
+    # do not create log in dry-run mode, just output to stdout/stderr
+    if options.dry_run:
+        old_stdout = 1
+        old_stderr = 2
+    else:
+        fd = os.open(logfile_dpkg, os.O_RDWR|os.O_CREAT, 0644)
+        old_stdout = os.dup(1)
+        old_stderr = os.dup(2)
+        os.dup2(fd,1)
+        os.dup2(fd,2)
+
+    try:
+        if (options.minimal_upgrade_steps or 
+            # COMPAT with the mispelling
+            apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps", False) or
+            apt_pkg.config.find_b("Unattended-Upgrade::MinimalSteps", False)):
+            open("/var/run/unattended-upgrades.pid", "w").write("%s" % os.getpid())
+            # try upgrade all "pkgs" in minimal steps
+            pkg_install_success = upgrade_in_minimal_steps(
+                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
+        else:
+            pkg_install_success = upgrade_normal(
+                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
+    except Exception as e:
+        # print unhandled exceptions here this way, while stderr is redirected
+        os.write(old_stderr, "Exception: %s" % e)
+
+    # restore
+    os.dup2(old_stdout, 1)
+    os.dup2(old_stderr, 2)
+
+    return pkg_install_success
+
 
 def _setup_alternative_rootdir(rootdir):
     # clear system unattended-upgrade stuff
@@ -557,7 +691,8 @@ def main(options, rootdir=""):
         sys.exit(1)
 
     # get a cache
-    cache = apt.Cache(rootdir=rootdir)
+    cache = UnattendedUpgradesCache(rootdir=rootdir, 
+                                    allowed_origins=allowed_origins)
     if cache._depcache.broken_count > 0:
         print _("Cache has broken packages, exiting")
         logging.error(_("Cache has broken packages, exiting"))
@@ -570,16 +705,27 @@ def main(options, rootdir=""):
     pkgs_kept_back = []
     pkgs_auto_removable = set([pkg.name for pkg in cache 
                                if pkg.is_auto_removable])
+
+    # now do the actual upgrade
     for pkg in cache:
         if options.debug and pkg.is_upgradable:
             logging.debug("Checking: %s (%s)" % (pkg.name, map(str, pkg.candidate.origins)))
         if (pkg.is_upgradable and 
-	    is_allowed_origin(pkg,allowed_origins)):
+	    is_allowed_origin(pkg.candidate, allowed_origins)):
             try:
                 pkg.mark_upgrade()
                 if check_changes_for_sanity(cache, allowed_origins,
                                             blacklisted_pkgs):
+                    # add to packages to upgrade
                     pkgs_to_upgrade.append(pkg)
+                    # re-eval pkgs_kept_back as the resolver may fail to 
+                    # directly upgrade a pkg, but that may work during
+                    # a subsequent operation, see debian bug #639840
+                    for pkgname in pkgs_kept_back:
+                        if (cache[pkgname].marked_install or
+                            cache[pkgname].marked_upgrade):
+                            pkgs_kept_back.remove(pkgname)
+                            pkgs_to_upgrade.append(cache[pkgname])
                 else:
                     logging.debug("sanity check failed")
                     rewind_cache(cache, pkgs_to_upgrade)
@@ -591,6 +737,7 @@ def main(options, rootdir=""):
                 pkgs_kept_back.append(pkg.name)
                 
 
+    pkgs_to_upgrade.sort(key=lambda p: p.name)
     pkgs = "\n".join([pkg.name for pkg in pkgs_to_upgrade])
     logging.debug("pkgs that look like they should be upgraded: %s" % pkgs)
            
@@ -628,9 +775,14 @@ def main(options, rootdir=""):
             if not item.is_trusted:
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
             if conffile_prompt(item.destfile):
-                # FIXME: skip package (means to re-run the whole marking again
+                # skip package (means to re-run the whole marking again
                 # and making sure that the package will not be pulled in by
-                # some other package again!
+                # some other package again!)
+                # 
+                # print to stdout to ensure that this message is part of
+                # the cron mail
+                print _("Package '%s' has conffile prompt and needs to be upgraded manually") % pkgname_from_deb(item.destfile)
+                # log to the logfile
                 logging.warning(_("Package '%s' has conffile prompt and needs to be upgraded manually") % pkgname_from_deb(item.destfile))
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
                 pkgs_kept_back.append(pkgname_from_deb(item.destfile))
@@ -677,6 +829,14 @@ def main(options, rootdir=""):
         logging.info(_("No packages found that can be upgraded unattended"))
         return
 
+    # check if its configured for install on shutdown, if so, the 
+    # environment UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN will
+    # be set by the unatteded-upgrades-shutdown script
+    if (not "UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN" in os.environ and
+        apt_pkg.config.find_b("Unattended-Upgrade::InstallOnShutdown", False)):
+        logger.debug("Configured to install on shutdown, so exiting now")
+        return
+
     # check if we are in dry-run mode
     if options.dry_run:
         logging.info("Option --dry-run given, *not* performing real actions")
@@ -686,46 +846,20 @@ def main(options, rootdir=""):
     pkgs = " ".join([pkg.name for pkg in pkgs_to_upgrade])
     logging.info(_("Packages that are upgraded: %s" % pkgs))
 
-    # set debconf to NON_INTERACTIVE, redirect output
-    os.putenv("DEBIAN_FRONTEND","noninteractive");
-    setup_apt_listchanges()
-    
-    # redirect to log
-    REDIRECT_INPUT = os.devnull
-    fd = os.open(REDIRECT_INPUT, os.O_RDWR)
-    os.dup2(fd,0)
-
+    # get log
     now = datetime.datetime.now()
     logfile_dpkg = os.path.join(
         _get_logdir(), 'unattended-upgrades-dpkg_%s.log' % now.isoformat('_'))
-    logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
-    fd = os.open(logfile_dpkg, os.O_RDWR|os.O_CREAT, 0644)
-    old_stdout = os.dup(1)
-    old_stderr = os.dup(2)
-    os.dup2(fd,1)
-    os.dup2(fd,2)
 
-    try:
+    # only perform install step if we actually have packages to install
+    pkg_install_success = None
+    shutdown_lock = -1
+    if len(pkgs_to_upgrade) > 0:
         # lock for the shutdown check - its fine if the system
         # is shutdown while downloading but not so much while installing
-        apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
-
-        if (options.minimal_upgrade_steps or 
-            apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps", False)):
-            open("/var/run/unattended-upgrades.pid", "w").write("%s" % os.getpid())
-            # try upgrade all "pkgs" in minimal steps
-            pkg_install_success = upgrade_in_minimal_steps(
-                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
-        else:
-            pkg_install_success = upgrade_normal(
-                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
-    except Exception, e:
-        # print unhandled exceptions here this way, while stderr is redirected
-        os.write(old_stderr, "Exception: %s" % e)
-
-    # restore
-    os.dup2(old_stdout, 1)
-    os.dup2(old_stderr, 2)
+        shutdown_lock = apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
+        # do install
+        pkg_install_success = do_install(cache, pkgs_to_upgrade, options, logfile_dpkg)
 
     # send a mail (if needed)
     if not options.dry_run:
@@ -735,6 +869,8 @@ def main(options, rootdir=""):
     # auto-reboot (if required and the config for this is set
     if (apt_pkg.config.find_b("Unattended-Upgrade::Automatic-Reboot", False) and
         os.path.exists(REBOOT_REQUIRED_FILE)):
+        if shutdown_lock > 0:
+            os.close(shutdown_lock)
         logging.warning("Found %s, rebooting" % REBOOT_REQUIRED_FILE)
         subprocess.call(["/sbin/reboot"])
         
